######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Apply local overrides to the current map.
Overrides defined in a xml document.

NOT IMPLEMENTED
"""


class map_override(object):
    """
    Apply local overrides to the current map. Overrides defined in a xml document.
    """
    def __init__(self, map_root, override_file, icvn, vriic, fic):
        pass

    def _set_value(self, map_root, path, variable, value):
        pass

    def _append_value(self, map_root, path, variable, value):
        pass

    def _reset_list(self, map_root, path, variable, value):
        pass
