"""
Validate and transform HIPAA X12 documents
"""

__author__ = "John Holland <john.holland@swmbh.org>"
__status__ = "production"
from .version import __version__
__date__ = "2015-02-13"
__revision__ = ""
